@@ -6,11 +6,12 @@
  *
  * Copyright (c) 1993 The Regents of the University of California.
  * Copyright (c) 1994-1997 Sun Microsystems, Inc.
+ * Copyright (c) 1998-1999 by Scriptics Corporation.
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * SCCS: @(#) tclAppInit.c 1.20 97/03/24 14:29:43
+ * RCS: @(#) $Id: tclAppInit.c,v 1.5 1999/04/16 00:48:03 stanton Exp $
  */
 
 #ifdef TCL_XT_TEST
@@ -24,26 +25,27 @@
  * Sun shared libraries to be used for Tcl.
  */
 
-#ifdef	__cplusplus
-extern "C" {
-#endif
+extern int matherr();
+int *tclDummyMathPtr = (int *) matherr;
 
-extern int matherr _ANSI_ARGS_((void));
-int (*tclDummyMathPtr) _ANSI_ARGS_((void)) = matherr;
 
 #ifdef TCL_TEST
+
+#include "tclInt.h"
+
+extern int		Procbodytest_Init _ANSI_ARGS_((Tcl_Interp *interp));
+extern int		Procbodytest_SafeInit _ANSI_ARGS_((Tcl_Interp *interp));
+extern int		TclObjTest_Init _ANSI_ARGS_((Tcl_Interp *interp));
 extern int		Tcltest_Init _ANSI_ARGS_((Tcl_Interp *interp));
-extern int		Unixtest_Init _ANSI_ARGS_((Tcl_Interp *interp));
-extern int		Objtest_Init _ANSI_ARGS_((Tcl_Interp *interp));
+#ifdef TCL_THREADS
+extern int		TclThread_Init _ANSI_ARGS_((Tcl_Interp *interp));
+#endif
+
 #endif /* TCL_TEST */
+
 #ifdef TCL_XT_TEST
-extern int		Xttest_Init _ANSI_ARGS_((Tcl_Interp *interp));
+extern int		Tclxttest_Init _ANSI_ARGS_((Tcl_Interp *interp));
 #endif
-
-#ifdef	__cplusplus
-}
-#endif
-
 
 /*
  *----------------------------------------------------------------------
@@ -63,18 +65,24 @@ extern int		Xttest_Init _ANSI_ARGS_((Tcl_Interp *interp));
  */
 
 int
-#ifdef _USING_PROTOTYPES_
-main (int    argc,		/* Number of command-line arguments. */
-     char **argv)		/* Values of command-line arguments. */
-#else
 main(argc, argv)
     int argc;			/* Number of command-line arguments. */
     char **argv;		/* Values of command-line arguments. */
-#endif
 {
+#ifdef TCL_TEST
+    /*
+     * Pass the build time location of the tcl library (to find init.tcl)
+     */
+    Tcl_Obj *path;
+    path = Tcl_NewStringObj(TCL_BUILDTIME_LIBRARY, -1);
+    TclSetLibraryPath(Tcl_NewListObj(1,&path));
+
+#endif
+
 #ifdef TCL_XT_TEST
     XtToolkitInitialize();
 #endif
+
     Tcl_Main(argc, argv, Tcl_AppInit);
     return 0;			/* Needed only to prevent compiler warning. */
 }
@@ -90,7 +98,7 @@ main(argc, argv)
  *
  * Results:
  *	Returns a standard Tcl completion code, and leaves an error
- *	message in interp->result if an error occurs.
+ *	message in the interp's result if an error occurs.
  *
  * Side effects:
  *	Depends on the startup script.
@@ -99,12 +107,8 @@ main(argc, argv)
  */
 
 int
-#ifdef _USING_PROTOTYPES_
-Tcl_AppInit (Tcl_Interp *interp)/* Interpreter for application. */
-#else
 Tcl_AppInit(interp)
     Tcl_Interp *interp;		/* Interpreter for application. */
-#endif
 {
     if (Tcl_Init(interp) == TCL_ERROR) {
 	return TCL_ERROR;
@@ -112,27 +116,28 @@ Tcl_AppInit(interp)
 
 #ifdef TCL_TEST
 #ifdef TCL_XT_TEST
-    if (Xttest_Init(interp) == TCL_ERROR) {
-	return TCL_ERROR;
-    }
-    Tcl_StaticPackage(interp, "Xttest", Xttest_Init,
-            (Tcl_PackageInitProc *) NULL);
+     if (Tclxttest_Init(interp) == TCL_ERROR) {
+	 return TCL_ERROR;
+     }
 #endif
     if (Tcltest_Init(interp) == TCL_ERROR) {
 	return TCL_ERROR;
     }
     Tcl_StaticPackage(interp, "Tcltest", Tcltest_Init,
             (Tcl_PackageInitProc *) NULL);
-    if (Objtest_Init(interp) == TCL_ERROR) {
+    if (TclObjTest_Init(interp) == TCL_ERROR) {
 	return TCL_ERROR;
     }
-    Tcl_StaticPackage(interp, "Objtest", Objtest_Init,
-            (Tcl_PackageInitProc *) NULL);
-    if (Unixtest_Init(interp) == TCL_ERROR) {
+#ifdef TCL_THREADS
+    if (TclThread_Init(interp) == TCL_ERROR) {
 	return TCL_ERROR;
     }
-    Tcl_StaticPackage(interp, "Unixtest", Unixtest_Init,
-            (Tcl_PackageInitProc *) NULL);
+#endif
+    if (Procbodytest_Init(interp) == TCL_ERROR) {
+	return TCL_ERROR;
+    }
+    Tcl_StaticPackage(interp, "procbodytest", Procbodytest_Init,
+            Procbodytest_SafeInit);
 #endif /* TCL_TEST */
 
     /*
