@@ -9,7 +9,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclUnixTime.c,v 1.2 1998/09/14 18:40:18 stanton Exp $
+ * RCS: @(#) $Id: tclUnixTime.c,v 1.4 1999/04/16 00:48:06 stanton Exp $
  */
 
 #include "tclInt.h"
@@ -165,12 +165,17 @@ TclpGetTimeZone (currentTime)
 #if defined(HAVE_TIMEZONE_VAR) && !defined (TCL_GOT_TIMEZONE)
 #   define TCL_GOT_TIMEZONE
     static int setTZ = 0;
+#ifdef TCL_THREADS
+    static Tcl_Mutex tzMutex;
+#endif
     int        timeZone;
 
+    Tcl_MutexLock(&tzMutex);
     if (!setTZ) {
         tzset();
         setTZ = 1;
     }
+    Tcl_MutexUnlock(&tzMutex);
 
     /*
      * Note: this is not a typo in "timezone" below!  See tzset
@@ -234,3 +239,61 @@ TclpGetTime(timePtr)
     timePtr->sec = tv.tv_sec;
     timePtr->usec = tv.tv_usec;
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclpGetDate --
+ *
+ *	This function converts between seconds and struct tm.  If
+ *	useGMT is true, then the returned date will be in Greenwich
+ *	Mean Time (GMT).  Otherwise, it will be in the local time zone.
+ *
+ * Results:
+ *	Returns a static tm structure.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+struct tm *
+TclpGetDate(time, useGMT)
+    TclpTime_t time;
+    int useGMT;
+{
+    CONST time_t *tp = (CONST time_t *)time;
+
+    if (useGMT) {
+	return gmtime(tp);
+    } else {
+	return localtime(tp);
+    }
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclpStrftime --
+ *
+ *	On Unix, we can safely call the native strftime implementation.
+ *
+ * Results:
+ *	The normal strftime result.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+size_t
+TclpStrftime(s, maxsize, format, t)
+    char *s;
+    size_t maxsize;
+    CONST char *format;
+    CONST struct tm *t;
+{
+    return strftime(s, maxsize, format, t);
+}
